"""
Test class for use with the unit test.
"""

from lily_unit_test.models.classification import Classification
from lily_unit_test.models.test_suite import TestSuite


class TestClassFail(TestSuite):

    CLASSIFICATION = Classification.FAIL

    def test_01_fail_by_return_false(self):
        return False

    def test_02_fail_by_exception(self):
        _a = 1 / 0


if __name__ == '__main__':

    TestClassFail().run()



"""
Test class for use with the unit test.
"""

from lily_unit_test.models.classification import Classification
from lily_unit_test.models.test_suite import TestSuite


class TestClassEmpty(TestSuite):
    CLASSIFICATION = Classification.FAIL


if __name__ == '__main__':

    TestClassEmpty().run()



"""
Test runner class.
Runs all test suites from a specific package folder (recursive)
"""

import inspect
import os
import sys
import webbrowser

from datetime import datetime
from lily_unit_test.models.html_report import generate_html_report
from lily_unit_test.models.logger import Logger
from lily_unit_test.models.test_suite import TestSuite


class TestRunner(object):

    REPORT_TIME_STAMP_FORMAT = "%Y%m%d_%H%M%S"
    REPORT_FOLDER = 'lily_unit_test_reports'

    ###########
    # Private #
    ###########

    @classmethod
    def _populate_test_suites(cls, test_suites_path):
        sys.path.append(test_suites_path)

        test_suites = []
        for current_folder, sub_folders, filenames in os.walk(test_suites_path):
            sub_folders.sort()
            for filename in filter(lambda x: x.endswith('.py'), filenames):
                import_path = os.path.join(current_folder[len(test_suites_path) + 1:], filename.replace('.py', ''))
                import_path = import_path.replace(os.sep, '.')
                module = __import__(str(import_path), fromlist=['*'])
                for attribute_name in dir(module):
                    attribute = getattr(module, attribute_name)
                    if inspect.isclass(attribute):
                        classes = inspect.getmro(attribute)
                        if len(classes) > 2 and TestSuite in classes:
                            test_suites.append(attribute)

        return test_suites

    @classmethod
    def _write_log_messages_to_file(cls, report_path, time_stamp, filename, logger):
        output_path = os.path.join(report_path, time_stamp)
        if not os.path.isdir(output_path):
            os.makedirs(output_path)

        with open(os.path.join(str(output_path), filename), 'w') as fp:
            fp.writelines(map(lambda x: '{}\n'.format(x), logger.get_log_messages()))

    ##########
    # Public #
    ##########

    @classmethod
    def run(cls, test_suites_path, options=None):
        test_run_result = False

        if options is None:
            options = {}

        test_suites_path = os.path.abspath(test_suites_path)

        report_path = options.get(
            'report_folder',
            os.path.join(os.path.dirname(test_suites_path), cls.REPORT_FOLDER)
        )
        write_log_files = not options.get('no_log_files', False)

        report_data = {}
        test_runner_log = Logger(False)
        time_stamp = datetime.now().strftime(cls.REPORT_TIME_STAMP_FORMAT)

        test_suites = cls._populate_test_suites(test_suites_path)

        include_filter = options.get('include_test_suites', [])
        if len(include_filter) > 0:
            test_suites = list(filter(lambda x: x.__name__ in include_filter, test_suites))

        exclude_filter = options.get('exclude_test_suites', [])
        if len(exclude_filter) > 0:
            test_suites = list(filter(lambda x: x.__name__ not in exclude_filter, test_suites))

        n_test_suites = len(test_suites)
        n_digits = len(str(n_test_suites))
        report_name_format = '{{:0{}d}}_{{}}'.format(n_digits)
        if n_test_suites > 0:
            n_test_suites_passed = 0
            test_runner_log.info('Run {} test suites from folder: {}'.format(n_test_suites, test_suites_path))

            for i, test_suite in enumerate(test_suites):
                test_suite_name = test_suite.__name__
                test_runner_log.empty_line()
                test_runner_log.info('Run test suite: {}'.format(test_suite_name))
                ts = test_suite()
                result = ts.run()
                if result is None or result:
                    n_test_suites_passed += 1
                    test_runner_log.info('Test suite {}: PASSED'.format(test_suite_name))
                else:
                    test_runner_log.error('Test suite {}: FAILED'.format(test_suite_name))

                report_id = report_name_format.format(i + 2, test_suite_name)
                report_data[report_id] = ts.log.get_log_messages()
                if write_log_files:
                    cls._write_log_messages_to_file(report_path, time_stamp, '{}.txt'.format(report_id), ts.log)

            test_runner_log.empty_line()
            ratio = 100 * n_test_suites_passed / n_test_suites
            test_runner_log.info('{} of {} test suites passed ({:.1f}%)'.format(
                                 n_test_suites_passed, n_test_suites, ratio))
            if n_test_suites == n_test_suites_passed:
                test_runner_log.info('Test runner result: PASSED')
                test_run_result = True
            else:
                test_runner_log.error('Test runner result: FAILED')

        else:
            test_runner_log.info('No test suites found in folder: {}'.format(test_suites_path))

        test_runner_log.shutdown()

        report_id = report_name_format.format(1, 'TestRunner')
        report_data[report_id] = test_runner_log.get_log_messages()
        if write_log_files:
            cls._write_log_messages_to_file(report_path, time_stamp, '{}.txt'.format(report_id), test_runner_log)

        if options.get('create_html_report', False):
            html_output = generate_html_report(report_data)
            filename = os.path.join(report_path, '{}_TestRunner.html'.format(time_stamp))
            if not os.path.isdir(report_path):
                os.makedirs(report_path)
            with open(filename, 'w') as fp:
                fp.write(html_output)

            if options.get('open_in_browser', False):
                webbrowser.open(filename)

        return test_run_result


if __name__ == '__main__':

    from lily_unit_test import test_classes

    test_options = {
        # 'report_folder': os.path.join(os.path.expanduser('~'), TestRunner.REPORT_FOLDER),
        'create_html_report': True,
        'open_in_browser': True,
        'no_log_files': True,
        # 'include_test_suites': ['TestClassEmpty', 'TestClassPass'],
        # 'exclude_test_suites': ['TestClassSetupFailException', 'TestClassTeardownFailException'],
    }

    test_result = TestRunner.run(os.path.dirname(test_classes.__file__), test_options)
    print('Test runner result:', test_result)



"""
Test suite class.
"""

from lily_unit_test.models.classification import Classification
from lily_unit_test.models.logger import Logger


class TestSuite(object):

    CLASSIFICATION = Classification.PASS

    def __init__(self):
        self.log = Logger()

    def run(self):
        test_suite_name = self.__class__.__name__
        self.log.info('Run test suite: {}'.format(test_suite_name))

        test_suite_result = False
        try:
            test_methods = sorted(list(filter(lambda x: x.startswith('test_'), dir(self))))
            n_tests = len(test_methods)
            assert n_tests > 0, 'No tests defined'

            # Run the setup
            try:
                setup_result = self.setup()
                if setup_result is not None and not setup_result:
                    self.log.error('Test suite {}: FAILED: setup failed'.format(test_suite_name))
                    setup_result = False
                else:
                    setup_result = True
            except Exception as e:
                self.log.error('Test suite {}: FAILED by exception in setup\nException: {}'.format(test_suite_name, e))
                setup_result = False

            if setup_result:
                n_passed = 0
                # Run the test methods
                for test_method in test_methods:
                    test_case_name = '{}.{}'.format(test_suite_name, test_method)
                    self.log.info('Run test case: {}'.format(test_case_name))
                    try:
                        method_result = getattr(self, test_method)()
                        if method_result is None or method_result:
                            n_passed += 1
                            self.log.info('Test case {}: PASSED'.format(test_case_name))
                        else:
                            self.log.error('Test case {}: FAILED'.format(test_case_name))

                    except Exception as e:
                        self.log.error('Test case {}: FAILED by exception\nException: {}'.format(test_case_name, e))

                ratio = 100 * n_passed / n_tests
                self.log.info('Test suite {}: {} of {} test cases passed ({:.1f}%)'.format(
                              test_suite_name, n_passed, n_tests, ratio))

                test_suite_result = n_passed == n_tests

            # Run the teardown
            try:
                self.teardown()
            except Exception as e:
                self.log.error('Test suite {}: FAILED by exception in teardown\nException: {}'.format(
                               test_suite_name, e))
                test_suite_result = False

        except Exception as e:
            self.log.error('Test suite {}: FAILED by exception\nException: {}'.format(test_suite_name, e))
            test_suite_result = False

        if self.CLASSIFICATION == Classification.FAIL:
            # We expect a failure
            test_suite_result = not test_suite_result
            if test_suite_result:
                self.log.info('Test suite failed, but accepted because classification is set to FAIL')
            else:
                self.log.error('Test suite passed, but a failure was expected because classification is set to FAIL')
        elif self.CLASSIFICATION != Classification.PASS:
            self.log.error('Test classification is not defined: {}'.format(self.CLASSIFICATION))
            test_suite_result = False

        if test_suite_result:
            self.log.info('Test suite {}: PASSED'.format(test_suite_name))
        else:
            self.log.error('Test suite {}: FAILED'.format(test_suite_name))

        self.log.shutdown()

        return test_suite_result

    ##############################
    # Override these when needed #
    ##############################

    def setup(self): return True
    def teardown(self): pass


if __name__ == '__main__':

    import os

    from lily_unit_test import test_classes
    from lily_unit_test.models.test_runner import TestRunner

    TestRunner.run(os.path.dirname(test_classes.__file__))



"""
Lily unit test package
"""

from lily_unit_test.models.classification import Classification
from lily_unit_test.models.test_runner import TestRunner
from lily_unit_test.models.test_suite import TestSuite

# For easy import:
Classification = Classification
TestRunner = TestRunner
TestSuite = TestSuite



"""
Generate HTML report.
"""

import os

from datetime import datetime
from string import Template
from lily_unit_test.models.logger import Logger


def generate_html_report(report_data):
    time_format = Logger.TIME_STAMP_FORMAT.split('.')[0]

    test_runner_start_message = ''
    test_runner_start = ''
    test_runner_end = ''
    test_runner_result = ''
    test_runner_result_message = ''
    test_suites_results = ''

    for key in report_data.keys():
        # Test runner result
        if '1_TestRunner' in key:
            # First message for start time and message
            parts = report_data[key][0].split('|')
            test_runner_start = parts[0].split('.')[0]
            test_runner_start_message = parts[-1].strip()

            # Second last message for result message
            test_runner_result_message = report_data[key][-2].split('|')[-1].strip()

            # Last message for end time and result
            test_runner_end = report_data[key][-1].split('|')[0].split('.')[0]
            if 'PASSED' in report_data[key][-1]:
                test_runner_result = 'PASSED'
            else:
                test_runner_result = 'FAILED'

        else:
            # Test suite results
            test_suites_results += '{}'.format(_generate_test_suite_results(key, report_data[key]))

    start = datetime.strptime(test_runner_start, time_format)
    end = datetime.strptime(test_runner_end, time_format)
    test_runner_duration = end - start

    template_filename = os.path.join(os.path.dirname(__file__), 'artifacts', 'html_report_template.html')
    with open(template_filename, 'r') as fp:
        template = fp.read()

    output = Template(template).substitute({
        'start_message': test_runner_start_message,
        'start_date': test_runner_start,
        'end_date': test_runner_end,
        'duration': test_runner_duration,
        'result': test_runner_result,
        'result_class': test_runner_result.lower(),
        'result_message': test_runner_result_message,
        'test_suites_results': test_suites_results
    })
    return output


def _generate_test_suite_results(test_suite_key, log_messages):
    time_format = Logger.TIME_STAMP_FORMAT.split('.')[0]

    test_name = test_suite_key.split('_')[-1]
    test_start = log_messages[0].split('|')[0].split('.')[0]
    test_end = log_messages[-1].split('|')[0].split('.')[0]
    if 'PASSED' in log_messages[-1]:
        test_result = 'PASSED'
    else:
        test_result = 'FAILED'

    start = datetime.strptime(test_start, time_format)
    end = datetime.strptime(test_end, time_format)
    duration = end - start

    output = '<div class ="test-suite {}">'.format(test_result.lower())
    output += '<span class="expand" title="Show/hide log messages" '
    output += 'id="button_{0}" onclick="show_log(\'{0}\')">&plus;</span> '.format(test_suite_key)
    output += '{}: {} ({})</div>\n'.format(test_name, test_result, duration)
    output += '<div class="log-messages" style="display:none" id="log_{}">\n'.format(test_suite_key)
    for log_message in log_messages:
        level = 'debug'
        parts = log_message.split('|')
        if len(parts) > 1:
            level = parts[1].strip().lower()
        if log_message.strip() == '':
            log_message = '&nbsp;'
        output += '<div class="log {}"><pre>{}</pre></div>\n'.format(level, log_message)
    output += '</div>\n'

    return output.strip()


if __name__ == '__main__':

    import time

    # Generate some report data
    dummy_report_data = {}

    # Test runner log messages
    tr_logger = Logger(log_to_stdout=False)
    tr_logger.info('Run 2 test suites from folder: C:\\path\\to\\test_suites')
    tr_logger.empty_line()
    tr_logger.info('Run test suite: TestCreateHtmlReport')

    # TestCreateHtmlReport log messages
    test_logger = Logger(log_to_stdout=False)
    test_logger.info('Run test suite: TestCreateHtmlReport')
    test_logger.info('Run test case: TestCreateHtmlReport.test_01_log_message_types')
    test_logger.debug('This is a debug message')
    test_logger.error('This is an error message')
    test_logger.debug('The next line is empty')
    test_logger.empty_line()
    test_logger.debug('This line contains HTML entities: > < " \' &. These must be escaped properly.')
    test_logger.handle_message(test_logger.TYPE_STDOUT, 'This is a stdout message\n')
    test_logger.handle_message(test_logger.TYPE_STDERR, 'This is a stderr message\n')
    test_logger.info('Test case TestCreateHtmlReport.test_01_log_message_types: PASSED')
    time.sleep(1.2)
    test_logger.info('Test suite TestPublishHtmlReport: 1 of 1 test cases passed (100.0%)')
    test_logger.info('Test suite TestCreateHtmlReport: PASSED')
    test_logger.shutdown()
    dummy_report_data['2_TestCreateHtmlReport'] = test_logger.get_log_messages()

    tr_logger.error('Test suite: TestCreateHtmlReport PASSED')
    tr_logger.empty_line()
    tr_logger.info('Run test suite: TestPublishHtmlReport')

    # TestPublishHtmlReport log messages
    test_logger = Logger(log_to_stdout=False)
    test_logger.info('Run test suite: TestPublishHtmlReport')
    test_logger.info('Run test case: TestPublishHtmlReport.test_01_upload_to_ftp')
    test_logger.error('Test case: TestPublishHtmlReport.test_01_upload_to_ftp: FAILED by exception')
    test_logger.error('Exception: connection refused by host,invalid authorisation')
    time.sleep(1.2)
    test_logger.info('Test suite TestPublishHtmlReport: 0 of 1 test cases passed (0.0%)')
    test_logger.error('Test suite TestClassFail: FAILED')
    test_logger.shutdown()
    dummy_report_data['3_TestPublishHtmlReport'] = test_logger.get_log_messages()

    tr_logger.info('Test suite: TestPublishHtmlReport FAILED')
    tr_logger.empty_line()
    tr_logger.info('1 of 2 test suites passed (50.0%)')
    tr_logger.error('Test runner result: FAILED')
    tr_logger.shutdown()
    dummy_report_data['1_TestRunner'] = tr_logger.get_log_messages()

    with open('test_report.html', 'w') as fp_out:
        fp_out.write(generate_html_report(dummy_report_data))



"""
Test class for use with the unit test.
"""

from lily_unit_test.models.test_suite import TestSuite


class TestClassPass(TestSuite):

    def test_01_pass_by_return_none(self):
        return None

    def test_02_pass_by_return_true(self):
        return True


if __name__ == '__main__':

    TestClassPass().run()



"""
Test class for use with the unit test.
"""

from lily_unit_test.models.classification import Classification
from lily_unit_test.models.test_suite import TestSuite


class TestClassSetupFailException(TestSuite):

    CLASSIFICATION = Classification.FAIL

    def setup(self):
        _a = 1 / 0

    def test_dummy(self):
        return True


if __name__ == '__main__':

    TestClassSetupFailException().run()



"""
Logger for the application.
"""

import sys
import threading
import time

from datetime import datetime


class Logger(object):

    TYPE_INFO = 'INFO'
    TYPE_DEBUG = 'DEBUG'
    TYPE_ERROR = 'ERROR'
    TYPE_STDOUT = 'STDOUT'
    TYPE_STDERR = 'STDERR'
    TYPE_EMPTY_LINE = 'EMPTY_LINE'

    TIME_STAMP_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
    _LOG_FORMAT = '{} | {:6} | {}'

    class _StdLogger(object):

        def __init__(self, logger, std_type):
            self._logger = logger
            self._type = std_type

        def write(self, message):
            self._logger.handle_message(self._type, message)

        def flush(self):
            pass

    def __init__(self, redirect_std=True, log_to_stdout=True):
        self._log_to_stdout = log_to_stdout
        self._log_messages = []
        self._output = ''

        self._orgStdout = sys.stdout
        self._orgStderr = sys.stderr
        if redirect_std:
            sys.stdout = self._StdLogger(self, self.TYPE_STDOUT)
            sys.stderr = self._StdLogger(self, self.TYPE_STDERR)

    def get_log_messages(self):
        return self._log_messages

    def shutdown(self):
        sys.stdout = self._orgStdout
        sys.stderr = self._orgStderr

    def info(self, message):
        self.handle_message(self.TYPE_INFO, '{}\n'.format(message))

    def debug(self, message):
        self.handle_message(self.TYPE_DEBUG, '{}\n'.format(message))

    def error(self, message):
        self.handle_message(self.TYPE_ERROR, '{}\n'.format(message))

    def empty_line(self):
        self.handle_message(self.TYPE_EMPTY_LINE, '')

    def handle_message(self, message_type, message_text):
        if message_type == self.TYPE_EMPTY_LINE:
            self._log_messages.append('')
            if self._log_to_stdout:
                self._orgStdout.write('\n')

        else:
            timestamp = datetime.now().strftime(self.TIME_STAMP_FORMAT)[:-3]
            self._output += message_text
            while '\n' in self._output:
                index = self._output.find('\n')
                line = self._LOG_FORMAT.format(timestamp, message_type, self._output[:index])
                self._output = self._output[index + 1:]
                self._log_messages.append(line)
                if self._log_to_stdout:
                    self._orgStdout.write('{}\n'.format(line))


if __name__ == "__main__":

    def _generate_error():
        def _exception(): _dummy = 1 / 0

        t = threading.Thread(target=_exception)
        t.start()
        time.sleep(1)


    test_logger = Logger()
    test_logger.info('This is an info message.')
    test_logger.debug('This is a debug message.')
    test_logger.error('This is an error message.')

    print('This is a stdout message.')
    print('This is a\nmulti line message.')

    _generate_error()

    test_logger.shutdown()

    print('\nMessages from logger')
    for log_message in test_logger.get_log_messages():
        print(log_message)
